package pilgrim_test

import (
	"flag"
	"fmt"
	"maps"
	"slices"
	"testing"

	"mvvasilev.dev/pilgrim/pilgrim"
)

func Test_CliContext_Url(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (string, bool) { return cli.Url() },
		pilgrim.Flag_Url,
		"driver://localhost:3306",
	)
}

func Test_CliContext_Driver(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (pilgrim.DbDriver, bool) { return cli.Driver() },
		pilgrim.Flag_Driver,
		"driver",
	)
}

func Test_CliContext_Username(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (string, bool) { return cli.Username() },
		pilgrim.Flag_Username,
		"username",
	)
}

func Test_CliContext_Password(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (string, bool) { return cli.Password() },
		pilgrim.Flag_Password,
		"password",
	)
}

func Test_CliContext_Host(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (string, bool) { return cli.Host() },
		pilgrim.Flag_Host,
		"localhost",
	)
}

func Test_CliContext_Port(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (string, bool) { return cli.Port() },
		pilgrim.Flag_Port,
		"3306",
	)
}

func Test_CliContext_Segments(t *testing.T) {
	expectedValue := []string{
		"segment1", "segment2", "segment3",
	}

	flag.Set(pilgrim.Flag_Segments, "segment1/segment2/segment3")

	value, _ := pilgrim.CliContext().Segments()

	Assert(
		t,
		slices.Equal(value, expectedValue),
		"Cli value '", pilgrim.Flag_Segments, "' differs from expected.", "Value:", value, "Expected:", expectedValue,
	)
}

func Test_CliContext_Args(t *testing.T) {
	expectedValue := map[string]string{
		"arg1": "val1",
		"arg2": "val2",
		"arg3": "val3",
	}

	flag.Set(pilgrim.Flag_Args, "arg1=val1&arg2=val2&arg3=val3")

	value, _ := pilgrim.CliContext().Args()

	Assert(
		t,
		maps.Equal(value, expectedValue),
		"Cli value '", pilgrim.Flag_Args, "' differs from expected.", "Value:", value, "Expected:", expectedValue,
	)
}

func Test_CliContext_Directory(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (string, bool) { return cli.Directory() },
		pilgrim.Flag_Directory,
		"./custom-migrations",
	)
}

func Test_CliContext_Script(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (string, bool) { return cli.Script() },
		pilgrim.Flag_Script,
		"2024-10-13/1_SomeScript.sql",
	)
}

func Test_CliContext_MigrationTable(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (string, bool) { return cli.MigrationTable() },
		pilgrim.Flag_Table,
		"CustomMigrationsTable",
	)
}

func Test_CliContext_MigrationTableSchema(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (string, bool) {
			return pilgrim.CliContext().MigrationTableSchema()
		},
		pilgrim.Flag_Schema,
		"public",
	)
}

func Test_CliContext_Up(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (bool, bool) { return cli.IsUp() },
		pilgrim.Flag_Up,
		true,
	)
}

func Test_CliContext_Down(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (bool, bool) { return cli.IsDown() },
		pilgrim.Flag_Down,
		true,
	)
}
func Test_CliContext_StrictOrdering(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (bool, bool) {
			return pilgrim.CliContext().IsStrictOrderingEnabled()
		},
		pilgrim.Flag_StrictOrdering,
		true,
	)
}

func Test_CliContext_DisableChecksumValidation(t *testing.T) {
	expectedValue := false

	flag.Set(pilgrim.Flag_DisableChecksumValidation, "true")

	value, _ := pilgrim.CliContext().IsChecksumValidationEnabled()

	Assert(
		t,
		value == expectedValue,
		"Cli value '", pilgrim.Flag_DisableChecksumValidation, "' differs from expected.", "Value:", value, "Expected:", expectedValue,
	)
}

func Test_CliContext_ValidateChecksums(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (bool, bool) {
			return cli.IsValidatingChecksums()
		},
		pilgrim.Flag_ValidateChecksums,
		true,
	)
}

func Test_CliContext_ValidateMigrationOrder(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (bool, bool) {
			return cli.IsValidatingMigrationOrder()
		},
		pilgrim.Flag_ValidateMigrationOrder,
		true,
	)
}

func Test_CliContext_ValidateLatest(t *testing.T) {
	test_cliValue(
		t,
		func(cli pilgrim.CliFlagRetriever) (bool, bool) { return cli.IsValidatingLatest() },
		pilgrim.Flag_ValidateLatest,
		true,
	)
}

func test_cliValue[T string | pilgrim.DbDriver | bool](t *testing.T, cliVal func(cli pilgrim.CliFlagRetriever) (value T, isProvided bool), flagName string, expectedVal T) {
	cli := pilgrim.CliContext()

	flag.Set(flagName, fmt.Sprintf("%v", expectedVal))

	cli.ParseFlags()

	value, _ := cliVal(cli)

	Assert(
		t,
		value == expectedVal,
		"Cli value '", flagName, "' differs from expected.", "Value:", value, "Expected:", expectedVal,
	)
}
