package pilgrim_test

import (
	"testing"

	"mvvasilev.dev/pilgrim/pilgrim"
	pilgrim_mock "mvvasilev.dev/pilgrim/pilgrim/test/mock"
)

const CliUrlValue = "postgres://cli_host:5432/cli_segment1/cli_segment2?cli_arg1=cli_val1&cli_arg2=cli_val2"
const CliDriverValue = "postgres"

func createPilgrimContextWithDependencies(
	cliFlags map[string]string,
	envVars map[string]string,
) (
	pilgrimContext pilgrim.PilgrimConfigurationRetriever,
	cliContext pilgrim.CliFlagRetriever,
	envContext pilgrim.EnvVarRetriever,
) {
	cliStringFlagFuncs := map[string]func() (value string, isProvided bool){}

	for k, v := range cliFlags {
		cliStringFlagFuncs[k] = func() (value string, isProvided bool) {
			return v, true
		}
	}

	cliContext = pilgrim_mock.MockCliContext(
		cliStringFlagFuncs,
		make(map[string]func() (value bool, isProvided bool)),
		func() (value []string, isProvided bool) { return []string{}, false },
		func() (value map[string]string, isProvided bool) { return map[string]string{}, false },
	)

	envVarStringFuncs := map[string]func() (value string, isProvided bool){}

	for k, v := range envVars {
		envVarStringFuncs[k] = func() (value string, isProvided bool) {
			return v, true
		}
	}

	envContext = pilgrim_mock.MockEnvVarContext(
		envVarStringFuncs,
		make(map[string]func() (value bool, isProvided bool)),
		func() (value []string, isProvided bool) { return []string{}, false },
		func() (value map[string]string, isProvided bool) { return map[string]string{}, false },
	)

	pilgrimContext = pilgrim.NewPilgrimContext(cliContext, envContext)

	return
}

func Test_PilgrimContext_CliDriverOverridesCliUrlDriver(t *testing.T) {
	context, _, _ := createPilgrimContextWithDependencies(
		map[string]string{
			pilgrim.Flag_Url:    CliUrlValue,
			pilgrim.Flag_Driver: CliDriverValue,
		},
		map[string]string{},
	)

	AssertEquals(t, context.UrlParts().Driver, CliDriverValue)
}

func Test_PilgrimContext_EnvDriverOverridesEnvUrlDriver(t *testing.T) {
	context, _, _ := createPilgrimContextWithDependencies(
		map[string]string{},
		map[string]string{
			pilgrim.EnvVarKey_Url:    "mariadb://env_host:3306/env_segment1/env_segment2?env_arg1=env_val1&env_arg2=env_val2",
			pilgrim.EnvVarKey_Driver: "postgres",
		},
	)

	AssertEquals(t, context.UrlParts().Driver, "postgres")
}

func Test_PilgrimContext_CliUrlDriverOverridesEnvDriver(t *testing.T) {
	context, _, _ := createPilgrimContextWithDependencies(
		map[string]string{
			pilgrim.Flag_Url: CliUrlValue,
		},
		map[string]string{
			pilgrim.EnvVarKey_Driver: "mariadb",
		},
	)

	AssertEquals(t, context.UrlParts().Driver, CliDriverValue)
}

func Test_PilgrimContext_CliDriverOverridesEnvDriver(t *testing.T) {
	context, _, _ := createPilgrimContextWithDependencies(
		map[string]string{
			pilgrim.Flag_Driver: CliDriverValue,
		},
		map[string]string{
			pilgrim.EnvVarKey_Driver: "mariadb",
		},
	)

	AssertEquals(t, context.UrlParts().Driver, CliDriverValue)
}

func Test_PilgrimContext_EnvDriverIsUsedIfNoCliDriverProvided(t *testing.T) {
	context, _, _ := createPilgrimContextWithDependencies(
		map[string]string{},
		map[string]string{
			pilgrim.EnvVarKey_Driver: "mariadb",
		},
	)

	AssertEquals(t, context.UrlParts().Driver, "mariadb")
}

func Test_PilgrimContext_EnvUrlDriverIsUsedIfNoCliDriverProvided(t *testing.T) {
	context, _, _ := createPilgrimContextWithDependencies(
		map[string]string{},
		map[string]string{
			pilgrim.EnvVarKey_Url: "mariadb://env_host:3306/env_segment1/env_segment2?env_arg1=env_val1&env_arg2=env_val2",
		},
	)

	AssertEquals(t, context.UrlParts().Driver, "mariadb")
}

func Test_PilgrimContext_EnvUrlDriverIsNotUsedIfNoCliDriverOverrideProvided(t *testing.T) {
	context, _, _ := createPilgrimContextWithDependencies(
		map[string]string{
			pilgrim.Flag_Driver: "postgres",
		},
		map[string]string{
			pilgrim.EnvVarKey_Url: "mariadb://env_host:3306/env_segment1/env_segment2?env_arg1=env_val1&env_arg2=env_val2",
		},
	)

	AssertEquals(t, context.UrlParts().Driver, "postgres")
}
